/**
 * Copyright (c) 2004-2009 Logos GmbH Rostock (www.logos-ib.de).
 * All rights reserved.
 *
 * THIS SOFTWARE IS  PROVIDED  "AS IS"  AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF  MERCHANTABILITY  AND  FITNESS  FOR  A PARTICULAR  PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL  LOGOS GMBH  OR ITS CONTRIBUTORS BE
 * LIABLE FOR ANY  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES  (INCL., BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR  PROFITS;
 * OR BUSINESS INTERRUPTION)  HOWEVER  CAUSED  AND  ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,  OR TORT (INCL.
 * NEGLIGENCE OR OTHERWISE)  ARISING  IN  ANY  WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package de.inovat.kex.bast.common;

/*
 * History:
 * 2009-03-30 [KS]  Erweiterung/Fehlebehebung getArchiveDataQueryResult()
 *                  bei der Archivanfrage ist zu bercksichtigen, dass mehrere
 *                  Versionen eines Konfigurationsobjekts unterschiedliche
 *                  Atg-Asp Kombinationen untersttzen knnen
 * 2008-03-28 [  ]  Suche nach letztem Datensatz: bei Gleichheit der Zeitstempel
 *                  wird zustzlich der Datenindex ausgewertet
 * 2008-03-27 [KS]  Bugfix getLastArchiveDataByArchiveTime und Erweiterung der
 *                  Suche falls gleiche Zeitstempel auftreten: "letzter"
 *                  Datensatz ist dann derjenige mit dem hhere Datenindex
 * 2007-07-26 [KS]  nderung Konfigurationsanfrage Objekte nach PIDs: weil
 *                  Datenzeitstempel vor Objekt-Gltigkeitsbereich liegen kann,
 *                  werden jetzt alle Objektversionen angefragt, die seit
 *                  dem Anfragebeginn gltig waren bzw. es jetzt sind
 * 2007-07-13 [KS]  NEU: Suchfunktion nach jngstem Datensatz (Datenzeitstempel)
 *                  UPDATE: Suchfunktion nach jngstem Datensatz (Archivzeit)
 */
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.Date;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.dav.daf.main.archive.ArchiveData;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKindCombination;
import de.bsvrz.dav.daf.main.archive.ArchiveDataQueryResult;
import de.bsvrz.dav.daf.main.archive.ArchiveDataSpecification;
import de.bsvrz.dav.daf.main.archive.ArchiveDataStream;
import de.bsvrz.dav.daf.main.archive.ArchiveOrder;
import de.bsvrz.dav.daf.main.archive.ArchiveQueryPriority;
import de.bsvrz.dav.daf.main.archive.ArchiveRequestOption;
import de.bsvrz.dav.daf.main.archive.ArchiveTimeSpecification;
import de.bsvrz.dav.daf.main.archive.TimingType;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * <p>stellt Funktionen zum Abfragen der Archivs bereit, die die Daten
 * f&uuml;r alle Versionen, der durch die &uuml;bergebenen Pids pr&auml;sentierenen
 * Objekte, bestimmt.</p>
 * <p>Einige Funktionen filtern die Daten nach dem Zeitstempel so, dass nur
 * die Datens&auml;tze zur&uuml;ckgegeben werden, die innerhalb der angegebenen
 * Zeitspanne liegen.</p>
 * 
 * @author LOGOS GmbH Rostock (www.logos-ib.de)
 * @author Ralf Trier
 * @version $Id: ArchiveRequestHelper.java 1431 2009-03-30 12:59:35Z Strey $
 */
abstract public class ArchiveRequestHelper {

	/** DebugLogger fuer Debug-Ausgaben. */
  private static Debug debug = Debug.getLogger();

  private ArchiveRequestHelper() {}

  private static final ArsDataComparator __ArsDataComparator = new
      ArsDataComparator();

  private static final ArchiveDataKindCombination DATA_KIND_COMBINATION =
      new ArchiveDataKindCombination(
          ArchiveDataKind.ONLINE,
          ArchiveDataKind.ONLINE_DELAYED,
          ArchiveDataKind.REQUESTED,
          ArchiveDataKind.REQUESTED_DELAYED);

  /**
   * gibt das Ergebnis f&uuml;r die ArchivAnfrage nach den SystemObjekten mit
   * der Pid im angegebenen Zeitraum <b>ungefiltert</b> zur&uuml;ck.
   *
   * @param dav ClientDavInterface
   * @param soPid String
   * @param atgPid String
   * @param aspPid String
   * @param von long
   * @param bis long
   * @return ArchiveDataQueryResult
   * @throws IOException
   */
  public static ArchiveDataQueryResult getArchiveDataQueryResult(
      ClientDavInterface dav, String soPid, String atgPid, String aspPid, long von, long bis)
      throws IOException
  {
    ArchiveTimeSpecification timeSpec = new ArchiveTimeSpecification(TimingType.DATA_TIME, false, von, bis);
//    logger.finer("getArchiveDataQueryResult(): timeSpec=%s", timeSpec );
    // (im Zeitraum) gltige Versionen des Systemobjekts ermitteln
    Collection<SystemObject> sos = dav.getDataModel().getObjects(soPid, 0, System.currentTimeMillis());
    TreeSet<SystemObject> sosSset = new TreeSet<SystemObject>(sos);
    List<ArchiveDataSpecification> specs = new ArrayList<ArchiveDataSpecification> (sos.size());
    for (SystemObject so : sosSset) {
      AttributeGroup atg = (so!=null) ? getAttributeGroup(so, atgPid) : null;
      Aspect asp = (atg!=null) ? getAspect(atg, aspPid) : null;
      DataDescription dataDescr = (asp!=null) ? new DataDescription(atg, asp, (short)0) : null;
      // nderungen am Objekttyp knnen dazu fhren, dass die gesuchte Atg:Asp
      // Kombination nicht an allen gelieferten Konfigurationsobjekten gltig
      // ist. In solchen Fllen kann/darf dieses Objekt nicht angefragt werden.
      if ((so!=null) && (dataDescr!=null)) {
      	debug.finer(String.format("getArchiveDataQueryResult(): so=%s %s %s:%s:%s %s:%s:%s",
      			so.getPid(),
      			so.getId(),
      			dataDescr.getAttributeGroup().getId(),
      			dataDescr.getAspect().getId(),
      			dataDescr.getSimulationVariant(),
      			dataDescr.getAttributeGroup().getPid(),
      			dataDescr.getAspect().getPid(),
      			dataDescr.getSimulationVariant()
      	));
      	ArchiveDataSpecification dataSpec =
      		new ArchiveDataSpecification(
      				timeSpec,
      				DATA_KIND_COMBINATION,
      				ArchiveOrder.BY_INDEX,
      				ArchiveRequestOption.NORMAL,
      				dataDescr,
      				so);

      	specs.add(dataSpec);
      } else {
      	debug.finer(String.format("getArchiveDataQueryResult(): so=%s " +
      			"verfgt nicht ber die gesuchten Daten %s:%s",
      			so, atgPid, aspPid));
      }
    }
    // dav.getArchive().setReceiveBufferSize(10*1024*1024);
    return dav.getArchive().request(ArchiveQueryPriority.MEDIUM, specs);
  }

  /**
   * gibt das Ergebnis f&uuml;r die ArchivAnfrage nach den SystemObjekten mit
   * der Pid im angegebenen Zeitraum <b>ungefiltert</b> als Stream zur&uuml;ck.
   *
   * @param dav ClientDavInterface
   * @param soPid String
   * @param atgPid String
   * @param aspPid String
   * @param von long
   * @param bis long
   * @return ArchiveDataStream
   * @throws IOException
   * @throws InterruptedException
   * @throws IllegalStateException
   */
  public static ArchiveDataStream getStream(ClientDavInterface dav, String soPid,
                                            String atgPid, String aspPid, long von,
                                            long bis) throws IOException, InterruptedException, IllegalStateException {

    ArchiveDataQueryResult arsResult = ArchiveRequestHelper.getArchiveDataQueryResult(dav, soPid, atgPid, aspPid, von, bis);
    return new MergedArchiveDataStream(arsResult);



  }

  /**
   * gibt die sortierten ArchivDaten im angegebenen Bereich (einschlie&szlig;lich
   * der Grenzen) <b>sortiert</b> und <b>gefiltert</b> zur&uuml;ck.
   *
   * @param dav ClientDavInterface
   * @param soPid String
   * @param atgPid String
   * @param aspPid String
   * @param von long
   * @param bis long
   * @throws IOException wenn Fehler beim Lesen aufgetreten sind.
   * @return Set
   */
  public static Set<ArchiveData> getArchivData(
      ClientDavInterface dav, String soPid, String atgPid, String aspPid, long von, long bis)
      throws IOException
  {
//    logger.info("Memory frei=%s max=%s total=%s", Runtime.getRuntime().freeMemory(), Runtime.getRuntime().maxMemory(), Runtime.getRuntime().totalMemory());
    TreeSet<ArchiveData> set = new TreeSet<ArchiveData> (__ArsDataComparator);
    ArchiveDataQueryResult arsResult = ArchiveRequestHelper.getArchiveDataQueryResult(dav, soPid, atgPid, aspPid, von, bis);
    try {
      ArchiveDataStream[] streams = arsResult.getStreams();
      for (ArchiveDataStream stream : streams) {
        ArchiveData arsData = stream.take();
        while (arsData != null) {
          if ( (arsData.getDataTime() <= bis) && (arsData.getDataTime() >= von)) {
            set.add(arsData);
          }
          arsData = stream.take();
        }
        // stream.abort();
      }
    }
    catch (IOException ex) {
      throw ex;
    }
    catch (IllegalStateException ex) {
      throw new IOException(ex.getMessage());
    }
    catch (InterruptedException ex) {
      throw new IOException(ex.getMessage());
    }
//    logger.info("Memory frei=%s max=%s total=%s", Runtime.getRuntime().freeMemory(), Runtime.getRuntime().maxMemory(), Runtime.getRuntime().totalMemory());
    // ars.archive.request.DataStream d;
    return set;
  }

  /**
   * bestimmt die AttributeGroup mit der Pid f&uuml;r das angegebene
   * SystemObject.
   *
   * @param so SystemObject
   * @param atgPid String
   * @return AttributeGroup
   */
  public static final AttributeGroup getAttributeGroup(SystemObject so,
      String atgPid) {
    List<AttributeGroup> atgList = so.getType().getAttributeGroups();
    for (AttributeGroup atg : atgList) {
      if (atg.getPid().equals(atgPid)) {
        return atg;
      }
    }
    return null;
  }

  /**
   * bestimmt den Aspect mit der Pid f&uuml;r die angegebene AttributeGroup.
   *
   * @param atg AttributeGroup
   * @param aspPid String
   * @return Aspect
   */
  public static final Aspect getAspect(AttributeGroup atg, String aspPid) {
    Collection<Aspect> aspList = atg.getAspects();
    for (Aspect asp : aspList) {
      if (asp.getPid().equals(aspPid)) {
        return asp;
      }
    }
    return null;
  }


  /**
   * <p>Bestimmung des zuletzt archivierten Datensatzes aus den im Archiv
   * vorliegenden Daten fr das gegebene Systemobjekt.
   * </p>
   * <p>Da das Archivsystem als Zustandsarchiv und nicht als Datenhaltung
   * ausgelegt ist, ist eine verlssliche Suche nur mit Brute-Force Methoden
   * mglich. Dazu werden alle nderungen bis zu einer maximalen Zeitspanne ab
   * aktuellem Zeitpunkt abgefragt und der zu letzt archivierte Datensatz
   * gesucht.<br>
   * Falls die Archivanfrage keine (gltigen) Ergebnisse enthlt, wird der Wert
   * <code>null</code> zurckgegeben.<br>
   * Ausgewertet werden alle Online, nachgelieferten und nachgeforderten Werte
   * von allen im Anfragezeitraum gltigen Versionen des Objekts.
   * </p>
   *
   * @param dav Datenverteilerverbindung
   * @param so das anzufragenden Objekt
   * @param descr Datenbeschreibung der gesuchten Daten
   * @param maxHistory long der maximale Zeitbereich in ms, bis zu dem vom
   *   aktuellen zeitpunkt aus nach einem Datensatz gesucht werden soll
   * @return der Archivdatensatz mit dem jngsten
   *   Archivzeitstempel im Anfragezeitraum
   * @throws IllegalStateException Falls die Archivanfrage nicht erfolgreich
   *   war und keine Ergebnisdatenstrme bestimmt werden knnen.
   */
  public static ArchiveData getLastArchiveDataByArchiveTime(
      ClientDavInterface dav,
      SystemObject so,
      DataDescription descr,
      long maxHistory)
  {
    ArchiveData result = null;
    long systemTime = System.currentTimeMillis();
    long queryTo = systemTime;
    long queryFrom = queryTo - maxHistory;
    ArchiveTimeSpecification timeSpec =
        new ArchiveTimeSpecification(
            TimingType.DATA_TIME, false, queryFrom, queryTo);
    try {
      // im Zeitraum gltige Versionen des Systemobjekts ermitteln und anfragen
      Collection<SystemObject> queryObjects =
          dav.getDataModel().getObjects(so.getPid(), queryFrom, systemTime);
      for (SystemObject queryObject : queryObjects) {
        ArchiveDataSpecification dataSpec =
            new ArchiveDataSpecification(
                timeSpec,
                DATA_KIND_COMBINATION,
                ArchiveOrder.BY_INDEX,
                ArchiveRequestOption.NORMAL,
                descr,
                queryObject);
        ArchiveDataQueryResult arsQueryResult =
            dav.getArchive().request(ArchiveQueryPriority.MEDIUM, dataSpec);
        for (ArchiveDataStream stream : arsQueryResult.getStreams()) {
          ArchiveData arsData = stream.take();
          while (arsData != null) {
            // Auswertung Archivantwort / Suche jngsten Archivzeitstempel
            if (arsData.getDataType().equals(DataState.DATA)) {
              if ((result == null)
               || (arsData.getArchiveTime() > result.getArchiveTime())
               || ((arsData.getArchiveTime() == result.getArchiveTime())
                && (arsData.getDataIndex()>result.getDataIndex())))
              {
                result = arsData;
              }
            }
            arsData = stream.take();
          }
        }
      }
    } catch (Exception ex) {
      debug.warning("getLastArchiveDataByArchiveTime(): Fehler bei Archivanfrage", ex);
      throw new IllegalStateException(ex);
    }

    debug.finer(" getLastArchiveDataByArchiveTime(" + so.getPid() + ':' + descr.getAttributeGroup().getPid() + ':' + descr.getAspect().getPid() + ") " +
                 "==> " + ((result == null) ? "null" : "data " + new Date(result.getDataTime())));

    return result;
  }

  /**
   * <p>Bestimmung des jngsten Datensatzes aus den im Archiv vorliegenden
   * Daten fr das gegebene Systemobjekt.
   * </p>
   * <p>Da das Archivsystem als Zustandsarchiv und nicht als Datenhaltung
   * ausgelegt ist, ist eine verlssliche Suche nur mit Brute-Force Methoden
   * mglich. Dazu werden alle nderungen bis zu einer maximalen Zeitspanne ab
   * aktuellem Zeitpunkt abgefragt und in den Antwortdaten der Datensatz mit
   * dem jngsten Datenzeitstempel gesucht. Daten die auerhalb des angefragten
   * Bereichs liegen werden ignoriert.<br>
   * Falls die Archivanfrage keine (gltigen) Ergebnisse enthlt, wird der Wert
   * <code>null</code> zurckgegeben.<br>
   * Ausgewertet werden alle Online, nachgelieferten und nachgeforderten Werte
   * von allen im Anfragezeitraum gltigen Versionen des Objekts.
   * </p>
   *
   * @param connection {@link ClientDavInterface} Datenverteilerverbindung
   * @param object {@link SystemObject} das anzufragenden Objekt
   * @param descr {@link DataDescription} Datenbeschreibung der gesuchten Daten
   * @param fromTime long Beginn Suchbereich
   * @param toTime long Ende Suchbereich
   * @return {@link ArchiveData} der Archivdatensatz mit dem jngsten
   *   Datenzeitstempel im Anfragezeitraum
   * @throws IllegalStateException Falls die Archivanfrage nicht erfolgreich
   *   war und keine Ergebnisdatenstrme bestimmt werden knnen.
   */
  public static ArchiveData getLastArchiveDataByDataTime(
      ClientDavInterface connection,
      SystemObject object,
      DataDescription descr,
      long fromTime, long toTime)
      throws java.lang.IllegalStateException
  {
    final long minHistory = IntervalHelper.TAG_IN_MS * 32; // mindestens 1 Monat durchsuchen
    ArchiveData result = null;
    long systemTime = System.currentTimeMillis();
    long queryTo = Math.max(toTime, systemTime);
    long queryFrom = Math.min(fromTime, systemTime - minHistory);

    ArchiveTimeSpecification timeSpec =
        new ArchiveTimeSpecification(TimingType.DATA_TIME, false, queryFrom, queryTo);
    ArchiveDataKindCombination dataKinds =
        new ArchiveDataKindCombination(
            ArchiveDataKind.ONLINE,
            ArchiveDataKind.ONLINE_DELAYED,
            ArchiveDataKind.REQUESTED,
            ArchiveDataKind.REQUESTED_DELAYED);

    try {
      // im Zeitraum gltige Versionen des Systemobjekts ermitteln und anfragen
      //
      // Weil Daten auch vor dem Beginn der Gltigkeit des Objekts eingespielt
      // werden knnen, muss mindestens bis zum aktuellen Zeitpunkt nach
      // gltigen Objekten gesucht werden !
      //
      Collection<SystemObject> queryObjects =
          connection.getDataModel().getObjects(object.getPid(), fromTime, systemTime);
      for (SystemObject queryObject : queryObjects) {
        ArchiveDataSpecification dataSpec =
            new ArchiveDataSpecification(
                timeSpec,
                dataKinds,
                ArchiveOrder.BY_INDEX,
                ArchiveRequestOption.NORMAL,
                descr,
                queryObject);

        debug.info(String.format("getLastArchiveDataByDataTime %s %2$td.%2$tm.%2$ty %2$tH:%2$tM %3$td.%3$tm.%3$ty %3$tH:%3$tM", queryObject.getPid(), new Date(queryFrom), new Date(queryTo)));

        ArchiveDataQueryResult arsQueryResult =
            connection.getArchive().request(ArchiveQueryPriority.MEDIUM, dataSpec);

        for (ArchiveDataStream stream : arsQueryResult.getStreams()) {
          ArchiveData arsData = stream.take();
          while (arsData != null) {
            // Auswertung Archivantwort / Suche jngsten Datensatz
            if ((arsData.getDataType().equals(DataState.DATA)) &&
                (arsData.getDataTime()>= fromTime) &&
                (arsData.getDataTime()<= toTime) ) {
              if ((result == null)
               || (arsData.getDataTime() > result.getDataTime())
               || ((arsData.getDataTime() == result.getDataTime())
                && (arsData.getDataIndex()>result.getDataIndex())))
              {
                result = arsData;
              }
            }
            arsData = stream.take();
          }
        }
      }
    } catch (Exception ex) {
      debug.warning("getLastArchiveDataByDataTime(): Fehler bei Archivanfrage", ex);
      throw new IllegalStateException(ex);
    }

    debug.finer("getLastArchiveDataByDataTime(" + object.getPid()+':'+descr.getAttributeGroup().getPid()+':'+descr.getAspect().getPid() +") " +
                 "==> " + ((result == null) ? "null" : "data " + new Date(result.getDataTime())));

    return result;
  }


  /**
   * Comparator zum Sortieren von ArchiveData nach DataTime, DataIndex.
   * @author unbekannt
   * @version 1.0
   */
  static class ArsDataComparator implements Comparator<ArchiveData> {

    public int compare(ArchiveData o1, ArchiveData o2) {
      if (o1.getDataTime() == o2.getDataTime()) {
        return (o1.getDataIndex() < o2.getDataIndex()) ? -1 : 1;
      }
      else {
        return (o1.getDataTime() < o2.getDataTime()) ? -1 : 1;
      }
    }

    public boolean equals(Object obj) {
      return (this.getClass() == obj.getClass());
    }
  }


  /**
   * ArchiveDataStream, der mehrere Streams zu einem Stream zusammenf&uuml;hrt.
   * @author unbekannt
   * @version 1.0
   */
  static class MergedArchiveDataStream implements ArchiveDataStream {

    ArchiveDataStream[] _Streams;
    int actStreamNr = 0;

    MergedArchiveDataStream(ArchiveDataQueryResult queryResult) throws
        IllegalStateException, InterruptedException {
        _Streams = queryResult.getStreams();
    }

    public ArchiveDataSpecification getDataSpecification() {
      return _Streams[actStreamNr].getDataSpecification();
    }

    public ArchiveData take() throws InterruptedException, IOException,
        IllegalStateException {
      if (actStreamNr < _Streams.length) {
        ArchiveData arsData = _Streams[actStreamNr].take();
        if (arsData == null) {
          actStreamNr++;
          return this.take();
        }
        return arsData;
      }
      return null;
    }

    public void abort() {
      for (ArchiveDataStream stream : _Streams) {
        if (stream != null) {
          stream.abort();
        }
      }
    }

  }

}
